S3_PROXY_BINARY_URLS = {
    "aarch64": "https://fast-s3-proxy.outerbounds.sh/linux-arm64/s3-proxy-0.1.1.gz",
    "x86_64": "https://fast-s3-proxy.outerbounds.sh/linux-amd64/s3-proxy-0.1.1.gz",
}

DEFAULT_PROXY_PORT = 8081
DEFAULT_PROXY_HOST = "localhost"
S3_PROXY_WRITE_MODES = ["origin-and-cache", "origin", "cache"]
